import sys
import random
import shutil
import pytest
import os.path as osp

import torch
import torch.nn.functional as F
from torch_geometric.nn import GraphSAGE
from torch_geometric.datasets import Planetoid

from torch_geometric.profile import profileit, timeit, get_stats_summary


@pytest.mark.skipif(not torch.cuda.is_available(), reason='CUDA not available')
def test_profile():
    root = osp.join('/', 'tmp', str(random.randrange(sys.maxsize)))

    dataset = Planetoid(root, 'PubMed')
    data = dataset[0].cuda()
    model = GraphSAGE(dataset.num_features, hidden_channels=64, num_layers=3,
                      out_channels=dataset.num_classes).cuda()
    optimizer = torch.optim.Adam(model.parameters(), lr=0.01)

    @profileit()
    def train(model, x, edge_index, y):
        model.train()
        optimizer.zero_grad()
        out = model(x, edge_index)
        loss = F.cross_entropy(out, y)
        loss.backward()
        return float(loss)

    @timeit()
    @torch.no_grad()
    def test(model, x, edge_index, y):
        model.eval()
        out = model(x, edge_index).argmax(dim=-1)
        return int((out == y).sum()) / y.size(0)

    stats_list = []
    for epoch in range(5):
        _, stats = train(model, data.x, data.edge_index, data.y)
        assert len(stats) == 6
        assert stats.time > 0
        assert stats.max_allocated_cuda > 0
        assert stats.max_reserved_cuda > 0
        assert stats.max_active_cuda > 0
        assert stats.nvidia_smi_free_cuda > 0
        assert stats.nvidia_smi_used_cuda > 0

        _, time = test(model, data.x, data.edge_index, data.y)
        assert time > 0

        if epoch >= 2:  # Warm-up
            stats_list.append(stats)

    stats_summary = get_stats_summary(stats_list)
    assert len(stats_summary) == 7
    assert stats_summary.time_mean > 0
    assert stats_summary.time_std > 0
    assert stats_summary.max_allocated_cuda > 0
    assert stats_summary.max_reserved_cuda > 0
    assert stats_summary.max_active_cuda > 0
    assert stats_summary.min_nvidia_smi_free_cuda > 0
    assert stats_summary.max_nvidia_smi_used_cuda > 0

    shutil.rmtree(root)
